#include <stdio.h>
#include <string.h>

#define BYTE 	char
#define WORD	short
#define DWORD	int

struct d88hdr_t {
	char title[17]; 
	BYTE rsrv[9];  
	BYTE protect;
	BYTE type;
	DWORD size;
	DWORD trkptr[164];
};

struct d88sct_t {
	BYTE cylinder, head, sector, sectorSize;
	WORD nsec;
	BYTE density, del, stat;
	BYTE rsrv[5];
	WORD size;
};

char *getType(BYTE type)
{
	switch(type)
	{
		case 0x00:
			return "2D";
			break;
		case 0x10:
			return "2DD";
			break;
		case 0x20:
			return "2HD";
			break;
	}
	return "UNKNOWN";
}

int main(int argc, char *argv[])
{
	FILE *file, *out;

	struct d88hdr_t header;
	int track = 0, numTracks= 0, totalSectors = 0;

	if(argc != 2)
	{
		printf("Please provide a d88 disk image\n");
		return 0;
	}	

	printf("Opening %s\n", argv[1]);
	file = fopen(argv[1], "rb");
	if(!file)
	{
		printf("Could not open file %s\n", argv[1]);
		return 0;
	}

	if(fread(&header, 1, sizeof(header), file) != sizeof(header))
	{
		printf("Could not read header\n");
		fclose(file);
		return(0);
	}

	out = fopen("out.raw", "wb");
	if(!out)
		printf("Could not open file out.raw\n");

	header.title[16] = '\0';
	printf("Title: %s\n", header.title);
	printf("Protected: %s\n", header.protect ? "yes" : "no");
	printf("Type: %s\n", getType(header.type));
	printf("Size: 0x%04X (%d)\n", header.size, header.size);

	if(header.trkptr[track] == 0x2B0)
		numTracks = 164;
	else
		numTracks = 160;

	for(track = 0; track < numTracks; track++)
	{
		if(header.trkptr[track] && fseek(file, header.trkptr[track], SEEK_SET) == 0)
		{
			struct	d88sct_t sctHeader;

			if(fread(&sctHeader, 1, sizeof(struct d88sct_t), file) == sizeof(struct d88sct_t))
			{
				int		sector = 0, numSectors = 0;
				BYTE	sectorData[1024];
	
				totalSectors++;
				printf("Track: %d Offset: 0x%04X\n  Cylinder: %d Head: %d, Sector: %d Size: %d\n",
					track, header.trkptr[track], sctHeader.cylinder, sctHeader.head, sctHeader.sector, 128 << sctHeader.sectorSize);

				printf("  Num Sectors: %d Density: %d  Deleted: %d Stat: 0x%02X Size: 0x%04X\n",
					sctHeader.nsec,
					sctHeader.density,
					sctHeader.del, sctHeader.stat, 
					sctHeader.size);

				if(fread(&sectorData, sizeof(BYTE), 128 << sctHeader.sectorSize, file) == 128 << sctHeader.sectorSize)
				{
					if(out)
						fwrite(sectorData, sizeof(BYTE), 128 << sctHeader.sectorSize, out);
				}
				else
				{
					printf("Error reading sector Data\n");
					return 0;
				}

				numSectors = sctHeader.nsec - 1;
				//printf("Data Size 0x%04X\n", sctHeader.nsec*(128 << sctHeader.sectorSize));
				for(sector = 0; sector < numSectors; sector ++)
				{
					if(fread(&sctHeader, 1, sizeof(struct d88sct_t), file) == sizeof(struct d88sct_t))
					{
						totalSectors++;
						printf("  Cylinder: %d Head: %d, Sector: %d Size: %d\n",
							sctHeader.cylinder, sctHeader.head, sctHeader.sector, 128 << sctHeader.sectorSize);
						printf("  Num Sectors: %d Density: %d  Deleted: %d Stat: 0x%02X Size: 0x%04X\n",
							sctHeader.nsec,
							sctHeader.density,
							sctHeader.del, 
							sctHeader.stat, 
							sctHeader.size);

						if(fread(&sectorData, sizeof(BYTE), 128 << sctHeader.sectorSize, file) == 128 << sctHeader.sectorSize)
						{
							if(out)
								fwrite(sectorData, sizeof(BYTE), 128 << sctHeader.sectorSize, out);
						}
						else
						{
							printf("Error reading sector Data\n");
							return 0;
						}
					}
				}
				
			}
		}
	}
	fclose(file);
	if(out)
		fclose(out);
	printf("Total Sectors: %d\n", totalSectors);
	return 0;
}